@@ -37,11 +37,14 @@ from .urls_patterns import URL_ADMIN_SERVER_INFO, URL_ADMIN_CLIENT_AUTHZ_RESOURC
     URL_ADMIN_REALMS, URL_ADMIN_USERS_COUNT, URL_ADMIN_FLOWS, URL_ADMIN_GROUP, URL_ADMIN_CLIENT_AUTHZ_SETTINGS, \
     URL_ADMIN_GROUP_MEMBERS, URL_ADMIN_USER_STORAGE, URL_ADMIN_GROUP_PERMISSIONS, URL_ADMIN_IDPS, \
     URL_ADMIN_USER_CLIENT_ROLES_AVAILABLE, URL_ADMIN_USERS, URL_ADMIN_CLIENT_SCOPES, \
-    URL_ADMIN_CLIENT_SCOPES_ADD_MAPPER, URL_ADMIN_CLIENT_SCOPE, URL_ADMIN_CLIENT_SECRETS
+    URL_ADMIN_CLIENT_SCOPES_ADD_MAPPER, URL_ADMIN_CLIENT_SCOPE, URL_ADMIN_CLIENT_SECRETS, \
+    URL_ADMIN_USER_REALM_ROLES
 
 
 class KeycloakAdmin:
 
+    PAGE_SIZE = 100
+
     def __init__(self, server_url, username, password, realm_name='master', client_id='admin-cli', verify=True, client_secret_key=None):
         """
 
@@ -120,6 +123,35 @@ class KeycloakAdmin:
     def token(self, value):
         self._token = value
 
+
+    def __fetch_all(self, url, query=None):
+        '''Wrapper function to paginate GET requests
+
+        :param url: The url on which the query is executed
+        :param query: Existing query parameters (optional)
+
+        :return: Combined results of paginated queries
+        '''
+        results = []
+
+        # initalize query if it was called with None
+        if not query:
+            query = {}
+        page = 0
+        query['max'] = self.PAGE_SIZE
+
+        # fetch until we can
+        while True:
+            query['first'] = page*self.PAGE_SIZE
+            partial_results = raise_error_from_response(
+                self.connection.raw_get(url, **query),
+                KeycloakGetError)
+            if not partial_results:
+                break
+            results.extend(partial_results)
+            page += 1
+        return results
+
     def import_realm(self, payload):
         """
         Import a new realm from a RealmRepresentation. Realm name must be unique.
@@ -147,13 +179,13 @@ class KeycloakAdmin:
 
     def create_realm(self, payload, skip_exists=False):
         """
-        Create a client
+        Create a realm
 
         ClientRepresentation: http://www.keycloak.org/docs-api/3.3/rest-api/index.html#_realmrepresentation
 
         :param skip_exists: Skip if Realm already exist.
         :param payload: RealmRepresentation
-        :return:  Keycloak server response (UserRepresentation)
+        :return:  Keycloak server response (RealmRepresentation)
         """
 
         data_raw = self.connection.raw_post(URL_ADMIN_REALMS,
@@ -168,8 +200,7 @@ class KeycloakAdmin:
         :return: users list
         """
         params_path = {"realm-name": self.realm_name}
-        data_raw = self.connection.raw_get(URL_ADMIN_USERS.format(**params_path), **query)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return self.__fetch_all(URL_ADMIN_USERS.format(**params_path), query)
 
     def get_idps(self):
         """
@@ -389,8 +420,7 @@ class KeycloakAdmin:
         :return: array GroupRepresentation
         """
         params_path = {"realm-name": self.realm_name}
-        data_raw = self.connection.raw_get(URL_ADMIN_GROUPS.format(**params_path))
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return self.__fetch_all(URL_ADMIN_GROUPS.format(**params_path))
 
     def get_group(self, group_id):
         """
@@ -423,7 +453,10 @@ class KeycloakAdmin:
                 return subgroup
             elif subgroup["subGroups"]:
                 for subgroup in group["subGroups"]:
-                    return self.get_subgroups(subgroup, path)
+                    result = self.get_subgroups(subgroup, path)
+                    if result:
+                        return result
+        # went through the tree without hits
         return None
 
     def get_group_members(self, group_id, **query):
@@ -436,8 +469,7 @@ class KeycloakAdmin:
         :return: Keycloak server response (UserRepresentation)
         """
         params_path = {"realm-name": self.realm_name, "id": group_id}
-        data_raw = self.connection.raw_get(URL_ADMIN_GROUP_MEMBERS.format(**params_path), **query)
-        return raise_error_from_response(data_raw, KeycloakGetError)
+        return self.__fetch_all(URL_ADMIN_GROUP_MEMBERS.format(**params_path), query)
 
     def get_group_by_path(self, path, search_in_subgroups=False):
         """
@@ -480,18 +512,6 @@ class KeycloakAdmin:
 
         :return: Http response
         """
-        name = payload['name']
-        path = payload['path']
-        exists = None
-
-        if name is None and path is not None:
-            path = "/" + name
-
-        elif path is not None:
-            exists = self.get_group_by_path(path=path, search_in_subgroups=True)
-
-        if exists is not None:
-            return str(exists)
 
         if parent is None:
             params_path = {"realm-name": self.realm_name}
@@ -501,8 +521,27 @@ class KeycloakAdmin:
             params_path = {"realm-name": self.realm_name, "id": parent, }
             data_raw = self.connection.raw_post(URL_ADMIN_GROUP_CHILD.format(**params_path),
                                                 data=json.dumps(payload))
+
         return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
 
+    def update_group(self, group_id, payload):
+        """
+        Update group, ignores subgroups.
+
+        :param group_id: id of group
+        :param payload: GroupRepresentation with updated information.
+
+        GroupRepresentation
+        http://www.keycloak.org/docs-api/3.2/rest-api/#_grouprepresentation
+
+        :return: Http response
+        """
+
+        params_path = {"realm-name": self.realm_name, "id": group_id}
+        data_raw = self.connection.raw_put(URL_ADMIN_GROUP.format(**params_path),
+                                           data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+
     def group_set_permissions(self, group_id, enabled=True):
         """
         Enable/Disable permissions for a group. Cannot delete group if disabled
@@ -646,6 +685,20 @@ class KeycloakAdmin:
                                             data=json.dumps(payload))
         return raise_error_from_response(data_raw, KeycloakGetError, expected_code=201, skip_exists=skip_exists)
 
+    def update_client(self, client_id, payload):
+        """
+        Update a client
+
+        :param client_id: Client id
+        :param payload: ClientRepresentation
+
+        :return: Http response
+        """
+        params_path = {"realm-name": self.realm_name, "id": client_id}
+        data_raw = self.connection.raw_put(URL_ADMIN_CLIENT.format(**params_path),
+                                           data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+
     def delete_client(self, client_id):
         """
         Get representation of the client
@@ -774,6 +827,23 @@ class KeycloakAdmin:
                                             data=json.dumps(payload))
         return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
 
+    def assign_realm_roles(self, user_id, client_id, roles):
+        """
+        Assign realm roles to a user
+
+        :param client_id: id of client (not client-id)
+        :param user_id: id of user
+        :param client_id: id of client containing role,
+        :param roles: roles list or role (use RoleRepresentation)
+        :return Keycloak server response
+        """
+
+        payload = roles if isinstance(roles, list) else [roles]
+        params_path = {"realm-name": self.realm_name, "id": user_id}
+        data_raw = self.connection.raw_post(URL_ADMIN_USER_REALM_ROLES.format(**params_path),
+                                            data=json.dumps(payload))
+        return raise_error_from_response(data_raw, KeycloakGetError, expected_code=204)
+
     def get_client_roles_of_user(self, user_id, client_id):
         """
         Get all client roles for a user.
