@@ -30,6 +30,7 @@ from .connection import ConnectionManager
 from .exceptions import raise_error_from_response, KeycloakGetError, \
     KeycloakRPTNotFound, KeycloakAuthorizationConfigError, KeycloakInvalidTokenError
 from .urls_patterns import (
+    URL_AUTH,
     URL_TOKEN,
     URL_USERINFO,
     URL_WELL_KNOWN,
@@ -156,9 +157,12 @@ class KeycloakOpenID:
 
         :return:
         """
-        return NotImplemented
+        params_path = {"authorization-endpoint": self.well_know()['authorization_endpoint'],
+                       "client-id": self.client_id,
+                       "redirect-uri": redirect_uri}
+        return URL_AUTH.format(**params_path)
 
-    def token(self, username="", password="", grant_type=["password"], code="", redirect_uri=""):
+    def token(self, username="", password="", grant_type=["password"], code="", redirect_uri="", totp=None, **extra):
         """
         The token endpoint is used to obtain tokens. Tokens can either be obtained by
         exchanging an authorization code or by supplying credentials directly depending on
@@ -172,12 +176,18 @@ class KeycloakOpenID:
         :param grant_type:
         :param code:
         :param redirect_uri
+        :param totp
         :return:
         """
         params_path = {"realm-name": self.realm_name}
         payload = {"username": username, "password": password,
                    "client_id": self.client_id, "grant_type": grant_type,
                    "code": code, "redirect_uri": redirect_uri}
+        if payload:
+            payload.update(extra)
+
+        if totp:
+            payload["totp"] = totp
 
         payload = self._add_secret_key(payload)
         data_raw = self.connection.raw_post(URL_TOKEN.format(**params_path),
